use super::*;

type Fwa<'a, A, E0, E1, Fallible> =
    Wrap<'a, Result<WrapE<'a, A, E0, Fallible>, E1>, <Fallible as MonadFailAny>::T>;

type Wwa<'a, A, E0, E1, Fallible> = WrapE<'a, WrapE<'a, A, E0, Fallible>, E1, Fallible>;

pub trait SpeculativeFail: MonadFailAny {
    fn _speculative_a_wb<'a, A: 'a, B: 'a, E0: 'a, E1: 'a>(
        a: Result<A, E0>,
        wb: Result<WrapE<'a, B, E0, Self>, E1>,
    ) -> WrapE<'a, (A, B), Result<E0, E1>, Self>
    where
        Self: 'a,
    {
        match wb {
            Ok(wb) => match a {
                Ok(a) => Self::map_err(<Self::W<E0> as Functor>::fmap(|b| (a, b), wb), Ok),
                Err(e0) => Self::fail(Ok(e0)),
            },
            Err(e1) => Self::fail(Err(e1)),
        }
    }

    fn _speculative_a_fwb<'a, A: 'a, B: 'a, E0: 'a, E1: 'a>(
        a: Result<A, E0>,
        fwb: Fwa<'a, B, E0, E1, Self>,
    ) -> WrapE<'a, (A, B), Result<E0, E1>, Self>
    where
        Self: 'a,
    {
        Self::stuff(<Self::T as Monad>::bind(fwb, |wb| {
            Self::unstuff(Self::_speculative_a_wb(a, wb))
        }))
    }

    fn _speculative_fa_wb<'a, A: 'a, B: 'a, E0: 'a>(
        fa: Wrap<'a, Result<A, E0>, Self::T>,
        wb: WrapE<'a, B, E0, Self>,
    ) -> WrapE<'a, (A, B), E0, Self>
    where
        Self: 'a,
    {
        <Self::W<E0> as ApplicativeTuple>::tuple((Self::stuff(fa), wb))
    }

    fn _speculative_wa_fwb<'a, A: 'a, B: 'a, E0: 'a, E1: 'a>(
        wa: WrapE<'a, A, E0, Self>,
        fwb: Fwa<'a, B, E0, E1, Self>,
    ) -> WrapE<'a, (A, B), Result<E0, E1>, Self>
    where
        Self: 'a,
    {
        Self::stuff(<<Self as MonadFailAny>::T as Monad>::join(
            <<Self as MonadFailAny>::T as Functor>::fmap(
                Self::unstuff,
                <Self::T as Functor>::fmap(
                    |selected| match selected {
                        Selected::A(a, fwb) => Self::_speculative_a_fwb(a, fwb),
                        Selected::B(fa, Ok(wb)) => {
                            Self::map_err(Self::_speculative_fa_wb(fa, wb), Ok)
                        }
                        Selected::B(_, Err(e1)) => Self::fail(Err(e1)),
                    },
                    <Self::T as ApplicativeSelect>::select(Self::unstuff(wa), fwb),
                ),
            ),
        ))
    }

    fn _speculative_fwa_wb<'a, A: 'a, B: 'a, E0: 'a, E1: 'a>(
        fwa: Fwa<'a, A, E0, E1, Self>,
        wb: WrapE<'a, B, E0, Self>,
    ) -> WrapE<'a, (A, B), Result<E0, E1>, Self>
    where
        Self: 'a,
    {
        <Self::W<Result<E0, E1>> as Functor>::fmap(
            |(b, a)| (a, b),
            Self::_speculative_wa_fwb(wb, fwa),
        )
    }

    fn speculative<'a, A: 'a, B: 'a, E0: 'a, E1: 'a>(
        wwa: Wwa<'a, A, E0, E1, Self>,
        wwb: Wwa<'a, B, E0, E1, Self>,
    ) -> WrapE<'a, (A, B), Result<E0, E1>, Self>
    where
        Self: 'a,
    {
        Self::stuff(<Self::T as Monad>::join(<Self::T as Functor>::fmap(
            Self::unstuff,
            <Self::T as Functor>::fmap(
                |selected| match selected {
                    Selected::A(Ok(wa), fwb) => Self::_speculative_wa_fwb(wa, fwb),
                    Selected::A(Err(e1), _) => Self::fail(Err(e1)),
                    Selected::B(fwa, Ok(wb)) => Self::_speculative_fwa_wb(fwa, wb),
                    Selected::B(_, Err(e1)) => Self::fail(Err(e1)),
                },
                <Self::T as ApplicativeSelect>::select(Self::unstuff(wwa), Self::unstuff(wwb)),
            ),
        )))
    }
}

impl<Fallible: ?Sized + MonadFailAny> SpeculativeFail for Fallible {}
